/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllib.gllib.modell.db;

import java.io.Serializable;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.List;

import de.bsvrz.iav.gllib.gllib.domain.ApproximationTyp;
import de.bsvrz.iav.gllib.gllib.domain.EreignisTyp;
import de.bsvrz.iav.gllib.gllib.domain.GanglinieTyp;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.Messwerte;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;

/**
 * Liste von Ganglinien als Blob serialisiert zum Speichern in der Datenbank.
 * Korrespondiert mit DAV-Attributliste <code>atl.ganglinie</code>.
 *
 * @author BitCtrl Systems GmbH, Thomas Thierfelder
 */
public class DbGanglinie implements Serializable {

    private static final long serialVersionUID = 2L;

    /**
     * PID des MQ.
     */
    private String mqPid;

    /**
     * ID des Ereignistyps.
     */
    private long ereignisTypId;

    /**
     * Ob es sich um eine Referenzganglinie handelt.
     */
    private boolean referenzGanglinie;

    /**
     * Zeitpunkt der letzten Verschmelzung.
     */
    private long letzteVerschmelzung;

    /**
     * Anzahl der Verschmelzungen insgesamt.
     */
    private long anzahlVerschmelzungen;

    /**
     * GL-Typ.
     */
    private int typ;

    /**
     * Approximationsverfahren.
     */
    private int approximationsverfahren;

    /**
     * B-Spline-Ordnung.
     */
    private int ordnung;

    /**
     * Liste aller Stuetzstellen.
     */
    private final ArrayList<DbStuetzstelle> stuetzstellen = new ArrayList<>();

    /**
     * Standardkonstruktor.
     */
    public DbGanglinie() {
        //
    }

    /**
     * Erfragt, ob es sich bei der uebergebenen Stuetzstelle um eine leere
     * Stuetzstelle handelt.
     *
     * @param value ein Q-/V-Wert einer Stuetzstelle.
     * @return ob es sich bei der uebergebenen Stuetzstelle um eine leere
     *         Stuetzstelle handelt.
     */
    public static final boolean isLeereStuetzstelle(final Double value) {
        return (value == null) || (Double.compare(value, Messwerte.UNDEFINIERT) == 0);
    }

    /**
     * Erfragt eine Instanz der innerhalb des GL-Lernen bzw. der Prognose benutzten
     * Klasse {@link MqGanglinie}, die aus den Daten dieses Objekts gespeist wird.
     *
     * @param messQuerschnitt der Messquerschnitt
     * @param ereignisTyp     der zugeordnete Ereignistyp
     * 
     * @return eine Instanz der innerhalb des GL-Lernen bzw. der Prognose benutzten
     *         Klasse {@link MqGanglinie}, die aus den Daten dieses Objekts gespeist
     *         wird, oder <code>null</code>, wenn die MQ-Ganglinie nicht
     *         rekonstruiert werden konnte (z.B., wenn der Ereignistyp im DAV nicht
     *         mehr vorhanden ist).
     */
    public final MqGanglinie convert(MessQuerschnitt messQuerschnitt, EreignisTyp ereignisTyp) {

        final MqGanglinie instanz = new MqGanglinie(messQuerschnitt, ereignisTyp);

        instanz.setAnzahlVerschmelzungen(anzahlVerschmelzungen);
        instanz.setApproximationDaK(ApproximationTyp.with(approximationsverfahren));
        instanz.setBSplineOrdnung(ordnung);

        instanz.setLetzteVerschmelzung(
                LocalDateTime.ofInstant(Instant.ofEpochMilli(letzteVerschmelzung), ZoneId.systemDefault()));
        instanz.setReferenz(referenzGanglinie);
        instanz.setTyp(GanglinieTyp.with(typ));
        for (final DbStuetzstelle stuetzStelle : stuetzstellen) {
            long zeitstempel;
            Double qKfz0, qLkw0, vPkw0, vLkw0;

            zeitstempel = stuetzStelle.getZeit();
            qKfz0 = stuetzStelle.getqKfz();

            if (isLeereStuetzstelle(qKfz0)) {
                qKfz0 = null;
            }
            qLkw0 = stuetzStelle.getqLkw();
            if (isLeereStuetzstelle(qLkw0)) {
                qLkw0 = null;
            }
            vPkw0 = stuetzStelle.getvPkw();
            if (isLeereStuetzstelle(vPkw0)) {
                vPkw0 = null;
            }
            vLkw0 = stuetzStelle.getvLkw();
            if (isLeereStuetzstelle(vLkw0)) {
                vLkw0 = null;
            }

            instanz.aktualisiereMesswert(zeitstempel, new Messwerte(qKfz0, qLkw0, vPkw0, vLkw0));
        }

        return instanz;
    }

    /**
     * Gibt die ID des Ereignistyps zurück.
     *
     * @return die ID des Ereignistyps.
     */
    public long getEreignisTypId() {
        return ereignisTypId;
    }

    /**
     * Legt die PID des Ereignistyps fest.
     *
     * @param ereignisTypId die ID des Ereignistyps.
     */
    public void setEreignisTypId(final long ereignisTypId) {
        this.ereignisTypId = ereignisTypId;
    }

    /**
     * Gibt die PID des MQs zurück.
     *
     * @return die PID des MQs.
     */
    public String getMqPid() {
        return mqPid;
    }

    /**
     * Legt die PID des MQs fest.
     *
     * @param mqPid die PID des MQs.
     */
    public void setMqPid(final String mqPid) {
        this.mqPid = mqPid;
    }

    /**
     * Flag, ob es sich um eine Referenzganglinie handelt.
     *
     * @return <code>true</code>, wenn die Ganglinie eine Referenzganglinie ist.
     */
    public boolean isReferenzGanglinie() {
        return referenzGanglinie;
    }

    /**
     * Flag, ob es sich um eine Referenzganglinie handelt.
     *
     * @param referenzGanglinie <code>true</code>, wenn die Ganglinie eine
     *                          Referenzganglinie ist.
     */
    public void setReferenzGanglinie(final boolean referenzGanglinie) {
        this.referenzGanglinie = referenzGanglinie;
    }

    /**
     * Gibt den Zeitpunkt der letzten Verschmelzung in Millisekunden zurück.
     *
     * @return der Zeitstempel der letzten Verschmelzung in Millisekunden.
     */
    public long getLetzteVerschmelzung() {
        return letzteVerschmelzung;
    }

    /**
     * Legt den Zeitpunkt der letzten Verschmelzung in Millisekunden fest.
     *
     * @param letzteVerschmelzung der Zeitstempel der letzten Verschmelzung in
     *                            Millisekunden.
     */
    public void setLetzteVerschmelzung(final long letzteVerschmelzung) {
        this.letzteVerschmelzung = letzteVerschmelzung;
    }

    /**
     * Gibt die Anzahl der Verschmelzungen insgesamt zurück.
     *
     * @return die Anzahl der Verschmelzungen.
     */
    public long getAnzahlVerschmelzungen() {
        return anzahlVerschmelzungen;
    }

    /**
     * Legt die Anzahl der Verschmelzungen insgesamt fest.
     *
     * @param anzahlVerschmelzungen die Anzahl der Verschmelzungen.
     */
    public void setAnzahlVerschmelzungen(final long anzahlVerschmelzungen) {
        this.anzahlVerschmelzungen = anzahlVerschmelzungen;
    }

    /**
     * Gibt den Typ der Ganglinie zurück.
     *
     * @return der Ganglinientyp.
     */
    public int getTyp() {
        return typ;
    }

    /**
     * Legt den Typ der Ganglinie fest.
     *
     * @param typ der Ganglinientyp.
     */
    public void setTyp(final int typ) {
        this.typ = typ;
    }

    /**
     * Gibt das Approximationsverfahren der Ganglinie zurück.
     *
     * @return das Approximationsverfahren der Ganglinie.
     */
    public int getApproximationsverfahren() {
        return approximationsverfahren;
    }

    /**
     * Legt das Approximationsverfahren der Ganglinie fest.
     *
     * @param approximationsverfahren das Approximationsverfahren der Ganglinie.
     */
    public void setApproximationsverfahren(final int approximationsverfahren) {
        this.approximationsverfahren = approximationsverfahren;
    }

    /**
     * Gibt die Ordnung des B-Splines zurück. Diese Eigenschaft ist nur relevant,
     * wenn als Approximationsverfahren B-Spline verwendet wird.
     *
     * @return die B-Spline-Ordnung.
     * @see #getApproximationsverfahren()
     */
    public int getOrdnung() {
        return ordnung;
    }

    /**
     * Legt die Ordnung des B-Splines fest. Diese Eigenschaft ist nur relevant, wenn
     * als Approximationsverfahren B-Spline verwendet wird.
     *
     * @param ordnung die B-Spline-Ordnung.
     * @see #getApproximationsverfahren()
     */
    public void setOrdnung(final int ordnung) {
        this.ordnung = ordnung;
    }

    /**
     * Gibt die Liste der Stützstellen der Ganglinie zurück.
     *
     * @return die Stützstellen der Ganglinie.
     */
    public List<DbStuetzstelle> getStuetzstellen() {
        return stuetzstellen;
    }

    /**
     * Legt die Liste der Stützstellen der Ganglinie fest.
     *
     * @param stuetzstellen die Stützstellen der Ganglinie.
     */
    public void setStuetzstellen(final List<DbStuetzstelle> stuetzstellen) {
        this.stuetzstellen.clear();
        this.stuetzstellen.addAll(stuetzstellen);
    }

    @Override
    public String toString() {
        String txt = getClass().getSimpleName();

        txt += "[";
        txt += "mqPid=" + mqPid;
        txt += ", ereignisTypId=" + ereignisTypId;
        txt += ", referenzGanglinie=" + referenzGanglinie;
        txt += ", letzteVerschmelzung=" + letzteVerschmelzung;
        txt += ", anzahlVerschmelzungen=" + anzahlVerschmelzungen;
        txt += ", typ=" + typ;
        txt += ", approximationsverfahren=" + approximationsverfahren;
        txt += ", ordnung=" + ordnung;
        txt += "]";

        return txt;
    }

}
