/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.Thread.UncaughtExceptionHandler;
import java.text.MessageFormat;
import java.time.LocalDateTime;
import java.time.LocalTime;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.Archiv;
import de.bsvrz.iav.gllib.gllib.domain.BetriebsMeldungsSender;
import de.bsvrz.iav.gllib.gllib.domain.EreignisKalender;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.ObjektFactory;
import de.bsvrz.iav.gllib.gllib.portals.DavArchiv;
import de.bsvrz.iav.gllib.gllib.portals.DavBetriebsmeldungsSender;
import de.bsvrz.iav.gllib.gllib.portals.DavChecker;
import de.bsvrz.iav.gllib.gllib.portals.DavEreignisKalender;
import de.bsvrz.iav.gllib.gllib.portals.DavGanglinienSpeicher;
import de.bsvrz.iav.gllib.gllib.portals.DavObjektFactory;
import de.bsvrz.iav.glprog.glprog.domain.GanglinienDatenbank;
import de.bsvrz.iav.glprog.glprog.domain.comm.GanglinienPrognoseAnfrageServer;
import de.bsvrz.iav.glprog.glprog.domain.comm.HistorischeGanglinienAnfrageServer;
import de.bsvrz.iav.glprog.glprog.domain.comm.OnlineDatenLeser;
import de.bsvrz.iav.glprog.glprog.portals.DavGanglinienAnfrageServer;
import de.bsvrz.iav.glprog.glprog.portals.DavGanglinienPrognoseAnfrageServer;
import de.bsvrz.iav.glprog.glprog.portals.DavOnlineDatenLeser;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

/**
 * Die Datenverteilerapplikation <i>Ganglinienprognose</i>. Segment 5
 * Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public final class GlProg implements StandardApplication, UncaughtExceptionHandler {

	/**
	 * Einsprungmethode der Applikation <i>Ganglinienprognose</i>.
	 *
	 * @param args die Kommandozeilenargumente.
	 * @see #parseArguments(ArgumentList)
	 */
	public static void main(final String[] args) {
		StandardApplicationRunner.run(new GlProg(), args);
	}

	/** Logger dieser Klasse. */
	private Debug log;

	private int maxAnfragen;
	private String dbSpeicherPid;
	private String dbVerzeichnis;

	private BetriebsMeldungsSender betriebsMeldungsSender;
	private OnlineDatenLeser onlineDatenLeser;
	private GanglinienPrognoseAnfrageServer prognoseAnfrageServer;
	private HistorischeGanglinienAnfrageServer historischeGanglinienAnfrageServer;

	private Archiv archiv;

	private EreignisKalender ereignisKalender;

	private ObjektFactory objektFactory;

	private DavGanglinienSpeicher ganglinienSpeicher;

	private ZyklischeAnfragenBearbeiter zyklischeAnfragenBearbeiter = new ZyklischeAnfragenBearbeiter();

	private LocalTime compressTime;

	public GlProg(ObjektFactory objektFactory, Archiv archiv, EreignisKalender ereignisKalender,
			OnlineDatenLeser onlineDatenLeser, GanglinienPrognoseAnfrageServer prognoseAnfrageServer,
			HistorischeGanglinienAnfrageServer historischeGanglinienAnfrageServer) {
		this.objektFactory = objektFactory;
		this.archiv = archiv;
		this.ereignisKalender = ereignisKalender;
		this.onlineDatenLeser = onlineDatenLeser;
		this.prognoseAnfrageServer = prognoseAnfrageServer;
		this.historischeGanglinienAnfrageServer = historischeGanglinienAnfrageServer;
		this.log = Debug.getLogger();
	}

	/**
	 * Initialisiert die Applikation.
	 */
	private GlProg() {

	}

	@Override
	public void initialize(final ClientDavInterface connection) throws Exception {
		MessageSender.getInstance()
				.setApplicationLabel(getClass().getSimpleName() + "_" + connection.getLocalApplicationObject().getId());

		DavChecker.checkConfiguration(connection);

		betriebsMeldungsSender = new DavBetriebsmeldungsSender(connection);

		log.fine("Initialisiere Objectfactory");
		objektFactory = new DavObjektFactory(connection);
		log.fine("Initialisiere Onlinedatenleser");
		onlineDatenLeser = new DavOnlineDatenLeser(connection);
		log.fine("Initialisiere Archiv");
		archiv = new DavArchiv(connection, betriebsMeldungsSender);
		log.fine("Initialisiere Ereigniskalender");
		ereignisKalender = new DavEreignisKalender((DavObjektFactory) objektFactory);
		SystemObject speicherObjekt = ((DavObjektFactory) objektFactory).getGanglinienSpeicherObjekt(dbSpeicherPid);
		log.fine("Initialisiere Ganglinienspeicher");
		ganglinienSpeicher = new DavGanglinienSpeicher((DavObjektFactory) objektFactory, speicherObjekt);

		/*
		 * Workaround, weil das Archiv die erste Anfrage der Applikation manchmal nicht
		 * beantworten will.
		 */
		connection.getArchive();

		log.fine("Initialisiere Ganglinienspeicher-Anfrageserver");
		historischeGanglinienAnfrageServer = new DavGanglinienAnfrageServer((DavObjektFactory) objektFactory,
				dbVerzeichnis, speicherObjekt, compressTime);
		log.fine("Initialisiere Ganglinienprognose-Anfrageserver");
		prognoseAnfrageServer = new DavGanglinienPrognoseAnfrageServer(this, (DavObjektFactory) objektFactory,
				maxAnfragen);

		final Thread mainThread = Thread.currentThread();
		Runtime.getRuntime().addShutdownHook(new Thread() {
			@Override
			public void run() {
				prognoseAnfrageServer.stop();
				historischeGanglinienAnfrageServer.stop();
				try {
					mainThread.join();
				} catch (@SuppressWarnings("unused") InterruptedException e) {
					// ignore
				}
			}
		});

		log.info(() -> "Ganglinienprognose bereit.");
	}

	/**
	 * Als Parameter wird unter anderen die Datenbank angegeben. Folgende Parameter
	 * kennt die Ganglinienprognose:
	 *
	 * <ul>
	 * <li><code>-dbIp=localhost</code>: Der Hostname oder die IP-Adresse der
	 * Derby-Datenbank, in der die Ganglinien gespeichert werden.</li>
	 * <li><code>-dbPort=1527</code>: Der Port der Derby-Datenbank, in der die
	 * Ganglinien gespeichert werden.</li>
	 * <li><code>-dbUser=derby</code>: Der Nutzername der Derby-Datenbank, in der
	 * die Ganglinien gespeichert werden.</li>
	 * <li><code>-dbPass=derby</code>: Das Nutzerpasswort der Derby-Datenbank, in
	 * der die Ganglinien gespeichert werden.</li>
	 * <li><code>-dbVerzeichnis=gldb</code>: Der Name der Derby-Datenbank, in der
	 * die Ganglinien gespeichert werden.</li>
	 * <li><code>-maxAnfragen=20</code>: Die maximale Anzahl von Anfragen die
	 * gleichzeitig beantwortet werden sollen. Kommen mehr Anfragen, werden diese in
	 * eine Warteschlange gelegt nacheinander abgearbeitet.</li>
	 * <li><code>-compressTime=23:00</code>: Der Zeitpunkt zu dem die tägliche
	 * Komprimierung der Ganglinien-Datenbank ausgeführt wird.</li>
	 * </ul>
	 */
	@Override
	public void parseArguments(final ArgumentList argumentList) throws Exception {
		Thread.setDefaultUncaughtExceptionHandler(this);
		maxAnfragen = Integer.parseInt(argumentList.fetchArgument("-maxAnfragen=20").asString()); //$NON-NLS-1$
		dbVerzeichnis = argumentList.fetchArgument("-dbVerzeichnis=gldb").asString();
		dbSpeicherPid = argumentList.fetchArgument("-speicherPid=").asString();
		String seconds = argumentList.fetchArgument("-compressTime=23:00").asString();
		compressTime = LocalTime.parse(seconds);
		log = Debug.getLogger();
	}

	/**
	 * Gibt die Exception im Log und als Betriebsmeldung aus und beendet
	 * anschließend die Applikation.
	 */
	@Override
	public void uncaughtException(final Thread t, final Throwable e) {
		String message = MessageFormat.format(
				"Es ist ein unbekannter Fehler im Thread {0} eingetreten. Die Applikation SWE Ganglinienprognose wird beendet. Stack Trace: {1}",
				t, getStackTrace(e));
		if (log == null) {
			System.err.println(message);
		} else {
			log.error(message);
			if (betriebsMeldungsSender != null) {
				betriebsMeldungsSender.fehler(null, message);
			}
		}
		System.exit(-1);
	}

	private String getStackTrace(final Throwable e) {
		final Writer result = new StringWriter();
		final PrintWriter printWriter = new PrintWriter(result);
		e.printStackTrace(printWriter);
		return result.toString();
	}

	public OnlineDatenLeser getOnlineDatenLeser() {
		return onlineDatenLeser;
	}

	public Archiv getArchiv() {
		return archiv;
	}

	public EreignisKalender getEreignisKalender() {
		return ereignisKalender;
	}

	public LocalDateTime getTime() {
		return objektFactory.getTime();
	}

	public DavGanglinienSpeicher getGanglinienSpeicher() {
		return ganglinienSpeicher;
	}

	public GanglinienDatenbank getGanglinienDatenbank() {
		return historischeGanglinienAnfrageServer.getSpeicher();
	}

	public void fehler(MessQuerschnitt mq, String message) {
		if (betriebsMeldungsSender != null) {
			betriebsMeldungsSender.fehler(mq, message);
		}
	}

	public void fehler(String message) {
		fehler(null, message);
	}

	public void warnung(MessQuerschnitt mq, String message) {
		if (betriebsMeldungsSender != null) {
			betriebsMeldungsSender.fehler(mq, message);
		}
	}

	public void warnung(String message) {
		warnung(null, message);
	}

	public ZyklischeAnfragenBearbeiter getZyklischeAnfragenBearbeiter() {
		return zyklischeAnfragenBearbeiter;
	}
}
