/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import de.bsvrz.iav.gllib.gllib.domain.EreignisTyp;
import de.bsvrz.iav.gllib.gllib.domain.PrognoseIntervall;

/**
 * Beschreibt ein Intervall im Prognosezeitraum. Ein Prognoseintervall wird
 * durch seinen Zeitraum und die im Zeitraum relevenaten Ganglinien beschrieben.
 * Relevant sind alle Ganglinien zu im Prognoseintervall anliegenden
 * Ereignissen.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public class PrognoseIntervallGanglinien {

    private static final class EreignisTypPrioritaetSorter implements Comparator<EreignisTyp>, Serializable {

        private static final long serialVersionUID = 1L;

        @Override
        public int compare(final EreignisTyp et1, final EreignisTyp et2) {

            if (et1.getPrioritaet() > et2.getPrioritaet()) {
                return -1;
            } else if (et1.getPrioritaet() < et2.getPrioritaet()) {
                return 1;
            }
            return 0;
        }
    }

    /** Der Zeitraum des Prognoseintervalls. */
    private final PrognoseIntervall intervall;

    private final List<EreignisTypGanglinien> ereignisTypGanglinien = new ArrayList<>();

    /**
     * Konstruiert das Prognoseintervall.
     *
     * @param intervall der Zeitraum des Prognoseintervalls.
     */
    public PrognoseIntervallGanglinien(final PrognoseIntervall intervall) {
        this.intervall = intervall;
    }

    /**
     * Gibt die im Intervall gültigen Ereignistypen zurück, zu denen absolute
     * Ganglinien existieren. Die Liste ist <em>absteigend</em> nach Priorität
     * sortiert.
     *
     * @return die sortierte Liste von Tagesereignissen.
     */
    public List<EreignisTyp> getEreignisTypenAbsolut() {
        final List<EreignisTyp> liste;

        // Alle Tagesereignisse sammeln ...
        liste = new ArrayList<>();
        for (final EreignisTypGanglinien eg : ereignisTypGanglinien) {
            if (eg.isTagesganglinie()) {
                liste.add(eg.getEreignisTyp());
            }
        }

        // ... und nach Priorität sortieren
        Collections.sort(liste, new EreignisTypPrioritaetSorter());

        return liste;
    }

    /**
     * Gibt die Liste der Ereignistypen zurück, zu denen nur relative Ganglinien
     * existieren.
     *
     * @return die Liste der Ereignistypen mit relativen Ganglinien.
     */
    public List<EreignisTypGanglinien> getEreignisTypGanglinienRelativ() {
        List<EreignisTypGanglinien> liste;

        liste = new ArrayList<>();
        for (final EreignisTypGanglinien eg : ereignisTypGanglinien) {
            if (!eg.isTagesganglinie()) {
                liste.add(eg);
            }
        }

        return liste;
    }

    /**
     * Gibt den Zeitraum des Prognoseintervall zurück.
     *
     * @return das Prognoseintervall.
     */
    public PrognoseIntervall getIntervall() {
        return intervall;
    }

    /**
     * Gibt die Tagesganglinie im Prognoseintervall zurück.
     *
     * @return die Tagesganglinie oder {@code null}, wenn keine existiert.
     */
    public EreignisTypGanglinien getTagesEreignisTypGanglinien() {
        int zaehler;
        EreignisTypGanglinien ganglinie;

        zaehler = 0;
        ganglinie = null;
        for (final EreignisTypGanglinien eg : ereignisTypGanglinien) {
            if (eg.isTagesganglinie()) {
                ganglinie = eg;
                zaehler++;
            }
        }

        assert (zaehler == 0) || (zaehler == 1) : "Es darf maximal eine Tagesganglinie geben"; //$NON-NLS-1$
        return ganglinie;
    }

    @Override
    public String toString() {
        String s;

        s = getClass().getName() + "["; //$NON-NLS-1$
        s += "intervall=" + intervall; //$NON-NLS-1$
        s += ", ereignisTypGanglinien=" + super.toString(); //$NON-NLS-1$

        return s + "]"; //$NON-NLS-1$
    }

    public List<EreignisTypGanglinien> getEreignisTypGanglinien() {
        return Collections.unmodifiableList(ereignisTypGanglinien);
    }

    public void addEreignistypGanglinie(EreignisTypGanglinien ereignisTypGanglinie) {
        ereignisTypGanglinien.add(ereignisTypGanglinie);
    }

    public void clear() {
        ereignisTypGanglinien.clear();
    }

    public void removeEreignisTypen(List<EreignisTyp> ereignisTypen) {

        Iterator<EreignisTypGanglinien> iterator = ereignisTypGanglinien.iterator();
        while (iterator.hasNext()) {
            EreignisTypGanglinien ganglinien;

            ganglinien = iterator.next();
            if (ereignisTypen.contains(ganglinien.getEreignisTyp())) {
                iterator.remove();
            }
        }

    }

}
