/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog.portals;

import java.time.LocalTime;
import java.util.List;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.GanglinieAnfrageTyp;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlAktualisierenAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlAnlegenAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlLesenAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlLoeschenAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlSchreibenAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlSpeicherAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlSpeicherAntwort;
import de.bsvrz.iav.gllib.gllib.portals.DavObjektFactory;
import de.bsvrz.iav.glprog.glprog.domain.GanglinienDatenbank;
import de.bsvrz.iav.glprog.glprog.domain.GanglinienSpeicherException;
import de.bsvrz.iav.glprog.glprog.domain.comm.GlUngueltigeAnfrage;
import de.bsvrz.iav.glprog.glprog.domain.comm.HistorischeGanglinienAnfrageServer;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Implementierung eines {@link HistorischeGanglinienAnfrageServer} zur Abfrage
 * von historischen Ganglinien über eine Datenverteiler-Verbindung.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 *
 */
public class DavGanglinienAnfrageServer implements HistorischeGanglinienAnfrageServer {

    private static final Debug LOGGER = Debug.getLogger();

    private DavGanglinienAntwortSender sender;
    private DavGanglinienAnfrageEmpfaenger empfaenger;
    private DerbyGanglinienDatenbank speicher;

    public DavGanglinienAnfrageServer(DavObjektFactory objektFactory, String dbVerzeichnis,
            SystemObject speicherObjekt, LocalTime compressTime) {
        LOGGER.fine("Initialisiere Datenbank");
        speicher = new DerbyGanglinienDatenbank(objektFactory, dbVerzeichnis, compressTime);
        LOGGER.fine("Installiere Antwortsender");
        sender = new DavGanglinienAntwortSender(objektFactory.getDav(), speicherObjekt);
        LOGGER.fine("Initialisiere Anfrageempfänger");
        empfaenger = new DavGanglinienAnfrageEmpfaenger(objektFactory, speicherObjekt);

        empfaenger.anfrageEmpfangen().addHandler(a -> beantworteAnfrage(a));
    }

    @Override
    public void beantworteAnfrage(GlSpeicherAnfrage anfrage) {

        GlSpeicherAntwort antwort;

        if (anfrage instanceof GlAktualisierenAnfrage) {
            antwort = erzeugeAktualisierenAntwort((GlAktualisierenAnfrage) anfrage);
        } else if (anfrage instanceof GlAnlegenAnfrage) {
            antwort = erzeugeAnlegenAntwort((GlAnlegenAnfrage) anfrage);
        } else if (anfrage instanceof GlLesenAnfrage) {
            antwort = erzeugeLesenAntwort((GlLesenAnfrage) anfrage);
        } else if (anfrage instanceof GlLoeschenAnfrage) {
            antwort = erzeugeLoeschenAntwort(anfrage);
        } else if (anfrage instanceof GlSchreibenAnfrage) {
            antwort = erzeugeSchreibenAntwort((GlSchreibenAnfrage) anfrage);
        } else {
            antwort = erzeugeFehlerAntwort(anfrage);
        }
        sender.anfrageBeantworten(antwort);
    }

    private GlSpeicherAntwort erzeugeAnlegenAntwort(GlAnlegenAnfrage anfrage) {
        GlSpeicherAntwort antwort = new GlSpeicherAntwort(anfrage.getTyp(), anfrage.getAnfrager(),
                anfrage.getAbsenderZeichen(), anfrage.getMessQuerschnitt());
        try {
            antwort.addGanglinien(speicher.erzeugeGanglinien(anfrage.getGanglinien()));
        } catch (GanglinienSpeicherException e) {
            antwort.setFehlerMeldung(e.getLocalizedMessage());
        }

        return antwort;
    }

    private GlSpeicherAntwort erzeugeAktualisierenAntwort(GlAktualisierenAnfrage anfrage) {
        GlSpeicherAntwort antwort = new GlSpeicherAntwort(anfrage.getTyp(), anfrage.getAnfrager(),
                anfrage.getAbsenderZeichen(), anfrage.getMessQuerschnitt());
        try {
            antwort.addGanglinien(speicher.aktualisiereGanglinien(anfrage.getGanglinien()));
        } catch (GanglinienSpeicherException e) {
            antwort.setFehlerMeldung(e.getLocalizedMessage());
        }

        return antwort;

    }

    private GlSpeicherAntwort erzeugeFehlerAntwort(GlSpeicherAnfrage anfrage) {
        String message = "Anfrage mit dem Typ " + anfrage.getTyp() + " konnte nicht interpretiert werden.";
        if (anfrage instanceof GlUngueltigeAnfrage) {
            message = ((GlUngueltigeAnfrage) anfrage).getMessage();
        }

        GlSpeicherAntwort antwort = new GlSpeicherAntwort(GanglinieAnfrageTyp.UNDEFINIERT, anfrage.getAnfrager(),
                anfrage.getAbsenderZeichen(), anfrage.getMessQuerschnitt());
        antwort.setFehlerMeldung(message);
        return antwort;
    }

    public GlSpeicherAntwort erzeugeLoeschenAntwort(GlSpeicherAnfrage anfrage) {
        GlSpeicherAntwort antwort = new GlSpeicherAntwort(anfrage.getTyp(), anfrage.getAnfrager(),
                anfrage.getAbsenderZeichen(), anfrage.getMessQuerschnitt());
        try {
            speicher.loescheGanglinien(anfrage.getMessQuerschnitt(), anfrage.getEreignisTypen(),
                    anfrage.getGanglinienId());
        } catch (GanglinienSpeicherException e) {
            antwort.setFehlerMeldung(e.getLocalizedMessage());
        }

        return antwort;
    }

    public GlSpeicherAntwort erzeugeSchreibenAntwort(GlSchreibenAnfrage anfrage) {
        GlSpeicherAntwort antwort = new GlSpeicherAntwort(anfrage.getTyp(), anfrage.getAnfrager(),
                anfrage.getAbsenderZeichen(), anfrage.getMessQuerschnitt());
        try {
            antwort.addGanglinien(speicher.schreibeGanglinien(anfrage.getMessQuerschnitt(), anfrage.getGanglinien()));
        } catch (GanglinienSpeicherException e) {
            antwort.setFehlerMeldung(e.getLocalizedMessage());
        }

        return antwort;
    }

    public GlSpeicherAntwort erzeugeLesenAntwort(GlLesenAnfrage anfrage) {
        GlSpeicherAntwort antwort = new GlSpeicherAntwort(anfrage.getTyp(), anfrage.getAnfrager(),
                anfrage.getAbsenderZeichen(), anfrage.getMessQuerschnitt());
        try {
            List<MqGanglinie> ganglinien = speicher.leseGanglinien(anfrage.getMessQuerschnitt(),
                    anfrage.getEreignisTypen());
            antwort.addGanglinien(ganglinien);
        } catch (GanglinienSpeicherException e) {
            antwort.setFehlerMeldung(e.getLocalizedMessage());
        }

        return antwort;
    }

    @Override
    public void stop() {
        speicher.close();
    }

    @Override
    public GanglinienDatenbank getSpeicher() {
        return speicher;
    }
}
