/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog.portals;

import java.time.Duration;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.NumberValue;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.VerkehrsDatenKurzZeitMq;
import de.bsvrz.iav.gllib.gllib.shared.Event;
import de.bsvrz.iav.glprog.glprog.GanglinienAuswahlmethode;
import de.bsvrz.iav.glprog.glprog.domain.GanglinienModellPrognose;
import de.bsvrz.iav.glprog.glprog.domain.VerkehrsDatenKurzZeitTrendExtraPolationMq;
import de.bsvrz.iav.glprog.glprog.domain.comm.OnlineDatenLeser;

/**
 * Implementierung eines {@link OnlineDatenLeser} zur Abfrage von Online-Daten
 * für die Ganglinienprognose über eine Datenverteiler-Verbindung.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class DavOnlineDatenLeser implements OnlineDatenLeser {

    private final ClientDavInterface dav;

    private final AttributeGroup atgVerkehrsDatenKurzZeitMq;
    private final Aspect aspAnalyse;

    private final AttributeGroup atgVerkehrsDatenKurzZeitTrendExtraPolationMq;
    private final Aspect aspPrognoseNormal;

    private final AttributeGroup atgGanglinienModellprognose;
    private final Aspect aspParameterSoll;

    private Event<MessQuerschnitt> ganglinienModellPrognoseChanged = new Event<>();

    public DavOnlineDatenLeser(ClientDavInterface dav) {
        this.dav = dav;

        DataModel dataModel = this.dav.getDataModel();

        atgVerkehrsDatenKurzZeitMq = dataModel.getAttributeGroup("atg.verkehrsDatenKurzZeitMq");
        aspAnalyse = dataModel.getAspect("asp.analyse");

        atgVerkehrsDatenKurzZeitTrendExtraPolationMq = dataModel
                .getAttributeGroup("atg.verkehrsDatenKurzZeitTrendExtraPolationMq");
        aspPrognoseNormal = dataModel.getAspect("asp.prognoseNormal");

        atgGanglinienModellprognose = dataModel.getAttributeGroup("atg.ganglinienModellPrognose");
        aspParameterSoll = dataModel.getAspect("asp.parameterSoll");

        List<SystemObject> mqs = dataModel.getType("typ.messQuerschnittAllgemein").getElements();
        dav.subscribeReceiver((r) -> triggerParameterChanges(r), mqs,
                new DataDescription(atgGanglinienModellprognose, aspParameterSoll), ReceiveOptions.normal(),
                ReceiverRole.receiver());
    }

    private Object triggerParameterChanges(ResultData[] results) {
        for (ResultData resultData : results) {
            SystemObject so = resultData.getObject();
            ganglinienModellPrognoseChanged.send(MessQuerschnitt.of(so.getPid()));
        }
        return null;
    }

    @Override
    public VerkehrsDatenKurzZeitMq getVerkehrsDatenKurzZeitMq(MessQuerschnitt mq) {

        if (atgVerkehrsDatenKurzZeitMq == null || aspAnalyse == null) {
            return VerkehrsDatenKurzZeitMq.invalid(dav.getTime());
        }

        SystemObject mqObj = dav.getDataModel().getObject(mq.getPid());
        if (mqObj == null) {
            return VerkehrsDatenKurzZeitMq.invalid(dav.getTime());
        }

        ResultData resultData = dav.getData(mqObj, new DataDescription(atgVerkehrsDatenKurzZeitMq, aspAnalyse), 0);
        if (!resultData.hasData()) {
            return VerkehrsDatenKurzZeitMq.invalid(resultData.getDataTime());
        }

        Data data = resultData.getData();

        Double vLkw = getDoubleValue(data.getItem("VLkw").getUnscaledValue("Wert"));
        Double vPkw = getDoubleValue(data.getItem("VPkw").getUnscaledValue("Wert"));
        Double qKfz = getDoubleValue(data.getItem("QKfz").getUnscaledValue("Wert"));
        Double qLkw = getDoubleValue(data.getItem("QLkw").getUnscaledValue("Wert"));

        return VerkehrsDatenKurzZeitMq.of(resultData.getDataTime(), vLkw, vPkw, qKfz, qLkw);
    }

    private Double getDoubleValue(NumberValue number) {
        if (number.isState()) {
            return null;
        }

        return number.doubleValue();
    }

    @Override
    public VerkehrsDatenKurzZeitTrendExtraPolationMq getVerkehrsDatenKurzZeitTrendExtraPolationMq(MessQuerschnitt mq) {

        if (atgVerkehrsDatenKurzZeitTrendExtraPolationMq == null || aspPrognoseNormal == null) {
            return VerkehrsDatenKurzZeitTrendExtraPolationMq.invalid(dav.getTime());
        }

        SystemObject mqObj = dav.getDataModel().getObject(mq.getPid());
        if (mqObj == null) {
            return VerkehrsDatenKurzZeitTrendExtraPolationMq.invalid(dav.getTime());
        }

        ResultData resultData = dav.getData(mqObj,
                new DataDescription(atgVerkehrsDatenKurzZeitTrendExtraPolationMq, aspPrognoseNormal), 0);
        if (!resultData.hasData()) {
            return VerkehrsDatenKurzZeitTrendExtraPolationMq.invalid(resultData.getDataTime());
        }

        Data data = resultData.getData();

        Double vLkw = getDoubleValue(data.getItem("VLkwP").getUnscaledValue("Wert"));
        Double vPkw = getDoubleValue(data.getItem("VPkwP").getUnscaledValue("Wert"));
        Double qKfz = getDoubleValue(data.getItem("QKfzP").getUnscaledValue("Wert"));
        Double qLkw = getDoubleValue(data.getItem("QLkwP").getUnscaledValue("Wert"));

        return VerkehrsDatenKurzZeitTrendExtraPolationMq.of(resultData.getDataTime(), vLkw, vPkw, qKfz, qLkw);
    }

    @Override
    public GanglinienModellPrognose getGanglinienModellPrognose(MessQuerschnitt messQuerschnitt) {
        if (atgGanglinienModellprognose == null || aspParameterSoll == null) {
            return GanglinienModellPrognose.KEIN_PARAMETER;
        }

        SystemObject mqObj = dav.getDataModel().getObject(messQuerschnitt.getPid());
        if (mqObj == null) {
            return GanglinienModellPrognose.KEIN_PARAMETER;
        }

        ResultData resultData = dav.getData(mqObj, new DataDescription(atgGanglinienModellprognose, aspParameterSoll),
                0);
        if (!resultData.hasData()) {
            return GanglinienModellPrognose.KEIN_PARAMETER;
        }

        Data data = resultData.getData();

        GanglinienAuswahlmethode auswahlMethode = GanglinienAuswahlmethode
                .with(data.getUnscaledValue("GLAuswahlMethode").intValue());
        Duration matchingHorizont = Duration.ofSeconds(data.getUnscaledValue("GLPatternMatchingHorizont").longValue());
        Duration matchingIntervall = Duration.ofSeconds(data.getUnscaledValue("GLMatchingIntervall").longValue());
        Duration matchingOffset = Duration.ofSeconds(data.getUnscaledValue("GLPatterMatchingOffset").longValue());
        int maximalerMatchingFehler = data.getUnscaledValue("GLMaximalerMatchingFehler").intValue();

        return GanglinienModellPrognose.of(auswahlMethode, matchingHorizont, matchingIntervall, matchingOffset,
                maximalerMatchingFehler);
    }

    @Override
    public Event<MessQuerschnitt> ganglinienModellPrognoseChanged() {
        return ganglinienModellPrognoseChanged;
    }
}
