/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog.portals;

import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.GanglinienAnfrager;
import de.bsvrz.iav.gllib.gllib.domain.Messwerte;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.Stuetzstelle;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlProgAntwort;
import de.bsvrz.iav.glprog.glprog.domain.comm.PrognoseAntwortSender;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * 
 * Implementierung des {@link PrognoseAntwortSender} zum Versenden von Antworten
 * auf Anfragen an die Ganglinienprognose über die Datenverteilerschnittstelle.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class DavPrognoseAntwortSender implements PrognoseAntwortSender {

    private static final Debug LOGGER = Debug.getLogger();

    private ClientSenderInterface sender = new ClientSenderInterface() {

        @Override
        public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
            return true;
        }

        @Override
        public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
            sendeBereit = state == ClientSenderInterface.START_SENDING;
        }
    };

    private ClientDavInterface dav;
    private AttributeGroup antwortAtg;
    private Aspect antwortAsp;

    private GanglinienAnfrager anfrager;

    private boolean sendeBereit;

    public DavPrognoseAntwortSender(ClientDavInterface dav, GanglinienAnfrager anfrager) {
        this.dav = dav;
        this.anfrager = anfrager;
        antwortAtg = dav.getDataModel().getAttributeGroup("atg.prognoseGanglinienAntwort");
        antwortAsp = dav.getDataModel().getAspect("asp.antwort");
        try {
            dav.subscribeSender(sender, dav.getDataModel().getObject(anfrager.getId()), new DataDescription(antwortAtg, antwortAsp),
                    SenderRole.sender());
        } catch (OneSubscriptionPerSendData e) {
            LOGGER.warning(e.getLocalizedMessage());
        }
    }

    @Override
    public void sendeAntwort(GlProgAntwort antwort) {
        try {
            dav.sendData(konvertiereAntwort(antwort));
        } catch (DataNotSubscribedException | SendSubscriptionNotConfirmed e) {
            LOGGER.warning(e.getLocalizedMessage());
        }
    }

    private ResultData konvertiereAntwort(GlProgAntwort antwort) {

        Data daten = dav.createData(antwortAtg);

        daten.getTextValue("AbsenderZeichen").setText(antwort.getAbsenderZeichen());

        Array feld = daten.getArray("PrognoseGanglinie");
        feld.setLength(antwort.getGanglinien().size());
        int i = 0;
        for (final MqGanglinie ganglinie : antwort.getGanglinien()) {
            getDaten(ganglinie, feld.getItem(i++));
        }

        return new ResultData(dav.getDataModel().getObject(antwort.getAnfrager().getId()),
                new DataDescription(antwortAtg, antwortAsp), dav.getTime(), daten);
    }

    private void getDaten(MqGanglinie ganglinie, final Data daten) {

        daten.getReferenceValue("Messquerschnitt").setSystemObjectPid(ganglinie.getMessQuerschnitt().getPid());
        daten.getUnscaledValue("GanglinienVerfahren").set(ganglinie.getApproximationDaK().getCode());
        daten.getUnscaledValue("Ordnung").set(ganglinie.getBSplineOrdnung());
        daten.getTimeValue("ZeitpunktPrognoseBeginn").setMillis(ganglinie.getIntervall().getStart());
        daten.getTimeValue("ZeitpunktPrognoseEnde").setMillis(ganglinie.getIntervall().getEnd());

        Array stuetzStellenArray = daten.getArray("Stützstelle");
        List<Stuetzstelle<Messwerte>> stuetzstellen = ganglinie.getStuetzstellen();
        stuetzStellenArray.setLength(stuetzstellen.size());
        for (int j = 0; j < stuetzStellenArray.getLength(); j++) {
            Data stuetzstellenItem = stuetzStellenArray.getItem(j);
            Stuetzstelle<Messwerte> stuetzstelle = stuetzstellen.get(j);

            stuetzstellenItem.getTimeValue("Zeit").setMillis(stuetzstelle.getZeitstempel());

            setAttribute(stuetzstellenItem, "QKfz", stuetzstelle.getWert().getQKfz());
            setAttribute(stuetzstellenItem, "QLkw", stuetzstelle.getWert().getQLkw());
            setAttribute(stuetzstellenItem, "VPkw", stuetzstelle.getWert().getVPkw());
            setAttribute(stuetzstellenItem, "VLkw", stuetzstelle.getWert().getVLkw());
        }
    }

    private void setAttribute(Data stuetzstellenItem, String name, Double wert) {
        if (wert == null) {
            stuetzstellenItem.getScaledValue(name).set(Messwerte.UNDEFINIERT);
        } else {
            stuetzstellenItem.getScaledValue(name).set(wert);
        }
    }

    @Override
    public boolean isSendeBereit() {
        return sendeBereit;
    }

    @Override
    public void abmelden() {
        dav.unsubscribeSender(sender, dav.getDataModel().getObject(anfrager.getId()), new DataDescription(antwortAtg, antwortAsp));
    }
}
