/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog.util;

import static de.bsvrz.iav.gllib.gllib.domain.GanglinienMQOperationen.auschneiden;

import java.text.MessageFormat;
import java.time.Duration;

import de.bsvrz.iav.gllib.gllib.domain.Archiv;
import de.bsvrz.iav.gllib.gllib.domain.Intervall;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.Messwerte;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.VerkehrsDatenKurzZeitMq;
import de.bsvrz.iav.gllib.gllib.domain.comm.PrognoseAnfrage;
import de.bsvrz.iav.gllib.gllib.shared.LocalDateTimeKonverter;
import de.bsvrz.iav.glprog.glprog.GlProg;
import de.bsvrz.iav.glprog.glprog.domain.GanglinienModellPrognose;
import de.bsvrz.iav.glprog.glprog.domain.VerkehrsDatenKurzZeitTrendExtraPolationMq;
import de.bsvrz.iav.glprog.glprog.domain.comm.OnlineDatenLeser;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Hilfsklasse für das Bestimmen der Analyseganglinie.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public final class AnalyseganglinieHelper {

	private final GlProg applikation;
	private final OnlineDatenLeser onlineDatenLeser;
	private final Archiv archiv;

	private final PrognoseAnfrage prognoseAnfrage;
	private GanglinienModellPrognose prognoseParameter;

	public AnalyseganglinieHelper(GlProg applikation, PrognoseAnfrage anfrage) {
		this.applikation = applikation;
		this.onlineDatenLeser = applikation.getOnlineDatenLeser();
		this.archiv = applikation.getArchiv();
		this.prognoseAnfrage = anfrage;
	}

	/**
	 * Gibt die Anfrage zurück, zu der die Analyseganglinie bestimmt werden soll.
	 *
	 * @return die Anfrage.
	 */
	public PrognoseAnfrage getAnfrage() {
		return prognoseAnfrage;
	}

	/**
	 * Gibt die Parameter der Prognose zurück, die zur Bestimmung der
	 * Analyseganglinien verwendet werden.
	 *
	 * @return der Prognoseparameter.
	 */
	public GanglinienModellPrognose getPrognoseParameter() {
		return prognoseParameter;
	}

	/**
	 * Legt die Parameter der Prognose fest, die zur Bestimmung der
	 * Analyseganglinien verwendet werden.
	 *
	 * @param prognoseParameter der Prognoseparameter.
	 */
	public void setPrognoseParameter(final GanglinienModellPrognose prognoseParameter) {
		this.prognoseParameter = prognoseParameter;
	}

	/**
	 * Erzeugt aus der Archivganglinie, den Kurzzeitdaten und den Daten der
	 * kurzfristigen Prognose die Analyseganglinie.
	 *
	 * @return die Analyseganglinie.
	 */
	public MqGanglinie getAnalyseganglinie() {
		final VerkehrsDatenKurzZeitMq kurzzeitdaten = getKurzzeitdaten();
		final VerkehrsDatenKurzZeitTrendExtraPolationMq trendprognose = getTrendprognose();
		final MqGanglinie archivGanglinie = getArchivdaten(prognoseAnfrage, prognoseParameter, kurzzeitdaten);

		final MqGanglinie analysegangline = archivGanglinie.erzeugeKopie();
		addKurzzeitdaten(analysegangline, kurzzeitdaten);
		addTrendprognose(analysegangline, trendprognose);

		return analysegangline;
	}

	/**
	 * Wenn es Kurzzeitdaten gibt, werden diese der Analyseganglinie hinzugefügt.
	 * 
	 * @param analysegangline die Analyseganglinie
	 * @param kurzzeitdaten   die neuen Kurzzeitdaten
	 */
	private static void addKurzzeitdaten(final MqGanglinie analysegangline,
			final VerkehrsDatenKurzZeitMq kurzzeitdaten) {
		if (!kurzzeitdaten.isValid()) {
			return;
		}

		final Messwerte messwert;
		final Double qKfz = kurzzeitdaten.getQKfz();
		final Double qLkw = kurzzeitdaten.getQLkw();
		final Double vPkw = kurzzeitdaten.getVPkw();
		final Double vLkw = kurzzeitdaten.getVLkw();

		messwert = new Messwerte(qKfz, qLkw, vPkw, vLkw);
		analysegangline.aktualisiereMesswert(kurzzeitdaten.getZeitstempel(), messwert);
	}

	/**
	 * Wenn es eine Trendprognose der Kurzzeitdaten gibt, werden diese der
	 * Analyseganglinie hinzugefügt.
	 * 
	 * @param analysegangline die Ganglinie
	 * @param trendprognose   die neuen Kurzzeitdaten
	 */
	private static void addTrendprognose(final MqGanglinie analysegangline,
			final VerkehrsDatenKurzZeitTrendExtraPolationMq trendprognose) {
		if (!trendprognose.isValid()) {
			return;
		}

		final Messwerte messwert;
		final Double qKfz = trendprognose.getQKfz();
		final Double qLkw = trendprognose.getQLkw();
		final Double vPkw = trendprognose.getVPkw();
		final Double vLkw = trendprognose.getVLkw();

		messwert = new Messwerte(qKfz, qLkw, vPkw, vLkw);
		analysegangline.aktualisiereMesswert(trendprognose.getZeitstempel(), messwert);
	}

	/**
	 * Fragt die aktuellen Kurzzeitdaten des Messquerschnitts vom Datenverteiler ab.
	 * Die Methode blockiert, bis die Kurzzeitdaten empfangen wurden.
	 *
	 * @return die aktuellen Kurzzeitdaten, aber niemals <code>null</code>.
	 */
	private VerkehrsDatenKurzZeitMq getKurzzeitdaten() {
		final MessQuerschnitt mq = prognoseAnfrage.getMessQuerschnitt();
		return onlineDatenLeser.getVerkehrsDatenKurzZeitMq(mq);
	}

	/**
	 * Fragt die aktuellen Trenddaten des Messquerschnitts vom Datenverteiler ab.
	 * Die Methode blockiert, bis die Trenddaten empfangen wurden.
	 *
	 * @return die aktuellen Trenddaten, aber niemals <code>null</code>.
	 */
	private VerkehrsDatenKurzZeitTrendExtraPolationMq getTrendprognose() {
		final MessQuerschnitt mq = prognoseAnfrage.getMessQuerschnitt();
		return onlineDatenLeser.getVerkehrsDatenKurzZeitTrendExtraPolationMq(mq);
	}

	/**
	 * Ruft die Archivdaten ab. Die Intervallgrenzen der Abfrage werden durch den
	 * Zeitstempel der Kurzzeitdaten und dem Prognoseparameter des Messquerschnitts
	 * bestimmt.
	 *
	 * <p>
	 * Die Synchronisierung dieser statischen Methode bewirkt, dass immer nur eine
	 * Archivanfrage gleichzeit stattfinden kann.
	 * 
	 * @param anfrage       die Anfrage
	 * @param parameter     die Parameter der Prognose
	 * @param kurzzeitdaten die aktuellen Kurzzeitdaten.
	 * @return die Ganglinie, die aus den Archivdaten gebildet wurde.
	 */
	private synchronized MqGanglinie getArchivdaten(final PrognoseAnfrage anfrage,
			final GanglinienModellPrognose parameter, final VerkehrsDatenKurzZeitMq kurzzeitdaten) {
		final MessQuerschnitt mq = anfrage.getMessQuerschnitt();
		final Debug log = Debug.getLogger();

		if (!archiv.isArchiveAvailable()) {
			log.error("Das Archiv ist nicht verfügbar.");
			applikation.fehler("Das Archiv ist nicht verfügbar.");
			return new MqGanglinie(mq);
		}

		// Anfrage zusammenstellen
		final Duration patternMatchingOffset = parameter.getPatternMatchingOffset();
		final Intervall intervall;
		if (kurzzeitdaten.isValid()) {
			// Archivdaten abrufen vom Zeitpunkt der Kurzzeitdaten bis
			// Matching-Offset in die Vergangenheit.
			intervall = Intervall.of(kurzzeitdaten.getZeitstempel() - patternMatchingOffset.toMillis(),
					kurzzeitdaten.getZeitstempel());
		} else {
			// Archivdaten abrufen vom Startzeitpunkt der Anfrage bis
			// Matching-Offset in die Vergangenheit.
			final long startPrognoseZeitraum = LocalDateTimeKonverter.toMillis(anfrage.getPrognoseZeitraum().getVon());
			intervall = Intervall.of(startPrognoseZeitraum - patternMatchingOffset.toMillis(), startPrognoseZeitraum);
		}

		MqGanglinie archivGl = archiv.bestimmeArchivGanglinie("asp.analyse", anfrage.getMessQuerschnitt(), intervall);
		if (archivGl.getStuetzstellen().size() > 0) {
			auschneiden(archivGl, intervall);
			log.finer(() -> MessageFormat.format("Analyseganglinie aus Archivdaten für {0}", archivGl));
		}

		return archivGl;
	}

}
